/*
        sizeof.c 1.0 for X11R6 & GNU C Compiler

        Copyright (C) 1999 Syllabub
        Maroontress Fast Software.
*/

#include <stdio.h>

#include "packet.h"
#include "sizeof.h"

static void
InsertPAD(FILE *fp)
{
    fprintf(fp, "\t" "s += (4 - ((s) %% 4)) %% 4;\n");
}

static int
IsCardinal(char *name)
{
    return (strcmp(name, "CARD8") == 0
	    || strcmp(name, "CARD16") == 0
	    || strcmp(name, "CARD32") == 0);
}

static void
CountAny(FILE *fp, char *name, char *member)
{
    if (IsCardinal(name)) {
	/* fprintf(fp, "\t" "s += sizeof(%s);\n", name); */
	fprintf(fp, "\t" "s += sizeof(p->%s);\n", member);
    }
    else if (strcmp(name, "XEVENT") == 0) {
	fprintf(fp, "\t" "s += sizeof(XEVENT);\n");
    }
    else {
	fprintf(fp, "\t" "s = Sizeof_%s(&(p->%s), s);\n", name, member);
    }
}

static void
CountArray(FILE *fp, char *name, char *member)
{
    fprintf(fp, "\t" "s += sizeof(CARD16);\n");
    InsertPAD(fp);
    fprintf(fp, "\t" "for (k = 0; k < p->%s.num; ++k)\n", member);
    if (IsCardinal(name)) {
	fprintf(fp, "\t\t" "s += sizeof(%s);\n", name);
    }
    else {
	fprintf(fp, "\t\t" "s = Sizeof_%s(p->%s.val[0], s);\n", name, member);
    }
}

static void
CountList(FILE *fp, char *name, char *member)
{
    fprintf(fp, "\t" "s += sizeof(CARD16);\n");
    if (strcmp(name, "CARD16") != 0)
	InsertPAD(fp);
    fprintf(fp, "\t"
	    "for (c = p->%s; c != NULL; c = c->next) {\n", member);
    if (IsCardinal(name)) {
	fprintf(fp, "\t\t" "s += sizeof(%s);\n", name);
    }
    else {
	fprintf(fp, "\t\t" "s = Sizeof_%s(c->data, s);\n", name);
    }
    fprintf(fp, "\t" "}\n");
}

static void
CountMulti(FILE *fp, char *name, char *member)
{
    fprintf(fp, "\t" "s += sizeof(CARD16);\n");
    if (strcmp(name, "CARD16") != 0)
	InsertPAD(fp);
    fprintf(fp, "\t" "for (k = 0; k < p->%s.num; ++k)\n", member);
    fprintf(fp, "\t\t" "s += sizeof(%s);\n", name);
}

void
PrintSizeof(FILE *fp, PacketList *pl)
{
    int n;
    Type *t;
    Packet *ptr;

    fprintf(fp,
#if 0
            "\n"
            "int\n"
            "Sizeof_STR(STR *p, int s)\n"
            "{\n"
	    "\t" "s += sizeof(CARD8);\n"
	    "\t" "s += p->len;\n"
	    "\t" "return (s);\n"
	    "}\n"
#endif
	    "\n"
            "int\n"
            "Sizeof_STRING(STRING *p, int s)\n"
            "{\n"
	    "\t" "s += sizeof(CARD16);\n"
	    "\t" "s += p->len;\n"
	    "\t" "s += (4 - ((s) %% 4)) %% 4;\n"
	    "\t" "return (s);\n"
	    "}\n");

    for (ptr = pl->top; ptr != NULL; ptr = ptr->next) {
	if (ptr->head == NULL)
	    continue;
	if (ptr->to_do != Write && ptr->to_do != Both)
	    continue;
	fprintf(fp,
		"\n"
		"int\n"
		"Sizeof_%s(%s *p, int s)\n"
		"{\n",
		ptr->type, ptr->type);
	n = 0;
	for (t = ptr->head; t != NULL; t = t->next)
	    n |= (1 << t->is_list);
	if (n > 1) {
	    if (n & (1 << Listof)) {
		fprintf(fp, "\t" "LIST *c;\n");
	    }
	    else {
		fprintf(fp, "\t" "int k;\n");
	    }
	    fprintf(fp, "\n");
	}
	
	for (t = ptr->head; t != NULL; t = t->next) {
	    if (strcmp(t->name, "PAD") == 0)
		InsertPAD(fp);
	    else if (t->is_list == Array)
		CountArray(fp, t->name, t->member);
	    else if (t->is_list == Listof)
		CountList(fp, t->name, t->member);
	    else if (t->is_list == Multi)
		CountMulti(fp, t->name, t->member);
	    else
		CountAny(fp, t->name, t->member);
	}
	fprintf(fp, "\t" "return (s);\n"
		"}\n");
    }
}
